"""Computes generalization error of Bayes decision function.

To provide context, Bayes decision function predicts wine quality using weather and
winemaking quality such that

:math:`\\begin{align}
\\mathbf{y} &= \\mathbf{Xw}, \\text { where} \\\\
\\mathbf{X} &= [\\mathbf{x}_1, \\mathbf{x}_2, \\mathbf{x}_1^2, \\mathbf{x}_2^2, \\mathbf{x}_1\\mathbf{x}_2], \\\\
\\mathbf{w} &= [0.3 ,  0.1 ,  0.07, -0.1 ,  0.1].
\\end{align}`

For more details, see whitepaper at
`sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error
from smltheory.generate_data import provide_weights, compute_rescaling_factor


def generate_bayes_features(data):
    """Generates features for Bayes decision function.

     Bayes decision function for wine quality was set to be the outcome a second-order
     polynomial model of weather and winemaking quality and the interaction between the
     first-order version of the predictors.

     Parameters
     -----------
     data : pandas.core.frame.DataFrame
        A dataframe created using `generate_mult_trunc_normal()` from the `generate_data`
        module.

     Returns
     -------
     bayes_features : pandas.core.frame.DataFrame
         A dataframe that contains features needed to compute Bayes decision function predictions

     See Also
     --------
     generate_data.create_covariance_matrix()
     generate_data.generate_mult_trunc_normal()

     Examples
     ---------
     >>> mu=[5, 7]
     >>> sd=[1.2, 1.7]
     >>> rho_weather_winemaking =  0.35
     >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
     #generate data set
     >>> sample_size_gen_error = 150
     >>> data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
     ... sample_size=sample_size_gen_error)
     #compute features for Bayes decision function
     >>> compute_bayes_features(data=data_gen_error)
          weather  winemaking_quality    weather  winemaking_quality          0
     0    5.214258            6.655854  27.188488           44.300390  34.705340
     1    5.459996            4.313195  29.811557           18.603650  23.550027
     2    4.813125            8.830114  23.166172           77.970922  42.500444
     3    5.499881            6.345226  30.248688           40.261895  34.897987
     4    5.235549            8.265799  27.410970           68.323429  43.275991
     ..        ...                 ...        ...                 ...        ...
     145  6.020267            7.603160  36.243618           57.808035  45.773053
     146  3.982080            7.635946  15.856965           58.307675  30.406952
     147  5.654110            9.818515  31.968960           96.403227  55.514961
     148  8.112048            6.618283  65.805322           43.801670  53.687829
     149  3.384190            6.122903  11.452739           37.489943  20.721065
     """

    # Feature columns
    predictors = data[["weather", "winemaking_quality"]]
    bayes_features = pd.concat(objs=[predictors, predictors**2, predictors.prod(axis=1)], axis=1)

    return bayes_features


def compute_bayes_intercept(data):
    """Computes intercept for Bayes decision function.

    Bayes decision function for wine quality was set to be the outcome a second-order
    polynomial model of weather and winemaking quality and the interaction between the
    first-order version of the predictors.

    Parameters
    ----------
    data: pandas.core.frame.DataFrame
       A pd.DataFrame created using `generate_mult_trunc_normal()` from the `generate_data`
       module

    Returns
    ---------
    intercept: np.array
        Intercept value of Bayes decision function.

    See Also
    ---------
    generate_data.generate_mult_trunc_normal()

    Examples
    ---------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data set
    >>> sample_size_gen_error = 150
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=sample_size_gen_error)
    #compute intercept of Bayes decision function
    >>> compute_bayes_intercept(data=data_gen_error)
    4.977118063990485
    """

    bayes_features = generate_bayes_features(data=data)

    # Intercept
    intercept = np.mean(data["wine_quality"]) - np.sum(provide_weights() * compute_rescaling_factor() *
                                                       np.mean(bayes_features, axis=0))

    return intercept


def compute_bayes_risk(data):
    """Computes generalization error/risk of Bayes decision function .

    Bayes decision function for wine quality was set to be the outcome a second-order
    polynomial model of weather and winemaking quality and the interaction between the
    first-order version of the predictors.

    Parameters
    --------
    data: pandas.core.frame.DataFrame
       A pd.DataFrame created using `generate_mult_trunc_normal()` from the `generate_data`
       module

    Returns
    --------
    bayes_risk: numpy.float64
        Generalization error/risk of Bayes decision function.

    See Also
    --------
    generate_data.generate_mult_trunc_normal()

    Examples
    --------
    >>> mu = [5, 7]
    >>> sd = [1.2, 1.7]
    >>> rho_weather_winemaking =  0.35
    >>> cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
    #generate data
    >>> sample_size_gen_error = 150
    >>> data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
    ... sample_size=sample_size_gen_error)
    >>> compute_bayes_risk(data=data_gen_error)
    0.25105984029122547
    """

    bayes_features = generate_bayes_features(data=data)

    # Compute intercept
    intercept = compute_bayes_intercept(data=data)

    # Compute predictions for wine quality
    test_pred = intercept + np.dot(a=bayes_features, b=provide_weights() * compute_rescaling_factor())

    # Compute mean squared error
    bayes_risk = mean_squared_error(y_true=data['wine_quality'], y_pred=test_pred)

    return bayes_risk



"""Gets data set(s) necessary for demonstrating supervised machine learning proposition.

Six data sets are used in this package to demonstrate supervised machine learning propositions:

1) `data_emp_loss`: data set for computing empirical loss.
2) `data_gen_error`: data set for computing generalization error.
3) `data_best_in_class`: data set for determining best function in constrained set (or class) of functions.
4) `data_est_error`: data set for depicting estimation error.
5) `data_opt_error`: data set for depicting optimization error.
6) `data_bv_tradeoff`: data set for depicting bias-variance tradeoff.

For more details, see
whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
"""

from importlib import resources
import pandas as pd


def get_data_emp_loss():
    """Returns data set for computing empirical loss  (i.e., "data_emp_loss.csv").

    Returns
    -------
    data_emp_loss:  pandas.core.frame.DataFrame
        Data set for computing empirical loss.

    Notes
    -----
    Data set for computing empirical loss was generated using the following code:

    .. code-block:: python

        sd = [1.5, 1.2]
        mu = [5, 7]
        sd = [1.2, 1.7]
        rho_weather_winemaking = 0.35
        cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
        # generate data
        sample_size_emp_loss = 150
        data_emp_loss = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                        mu=mu, sample_size=sample_size_emp_loss, seed=17)

    """
    with resources.path("smltheory.data", "data_emp_loss.csv") as path:
        data_emp_loss = pd.read_csv(path)

    return data_emp_loss


def get_data_gen_error():
    """Returns data set for computing generalization error (i.e., "data_gen_error.csv").

    Returns
    -------
    data_gen_error:  pandas.core.frame.DataFrame
        Data set for computing generalization error.

    Notes
    -----
    Data set for computing generalization error was generated using the following code:

    .. code-block:: python

        sd = [1.5, 1.2]
        mu = [5, 7]
        sd = [1.2, 1.7]
        rho_weather_winemaking = 0.35
        cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
        # generate data
        sample_size_gen_error= 1e4
        data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, mu=mu,
                         sample_size=sample_size_gen_error, seed = 21)

    """
    with resources.path("smltheory.data", "data_gen_error.csv") as path:
        data_gen_error = pd.read_csv(path)

    return data_gen_error


def get_data_best_in_class():
    """Returns data set for obtaining the constrained empirical risk minimizer (i.e., best function in 
    constrained set of functions). 

    Returns
    -------
    data_gen_error:  pandas.core.frame.DataFrame
        Data set for computing generalization error.

    Notes
    -----
    Data set for obtaining the constrained empirical risk minimizer (i.e., best function in
    constrained set/class of functions) was generated using the following code:

    .. code-block:: python

        sd = [1.5, 1.2]
        mu = [5, 7]
        sd = [1.2, 1.7]
        rho_weather_winemaking = 0.35
        cov_matrix = create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)
        # generate data
        sample_size_best_in_class = 1e4
        data_best_in_class = generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                             mu=mu, sample_size=sample_size_best_in_class, seed = 7)

    """
    with resources.path("smltheory.data", "data_best_in_class.csv") as path:
        data_best_in_class = pd.read_csv(path)

    return data_best_in_class


def get_data_est_error():
    """Returns data set for depicting estimation error.

    Returns
    -------
    data_est_error:  pandas.core.frame.DataFrame
        Data set for depicting estimation error

    Notes
    -----
    Data set for depicting estimation error was generated using the following code:

    .. code-block:: python

        sd = [1.5, 1.2]
        mu = [5, 7]
        sd = [1.2, 1.7]
        rho_weather_winemaking = 0.35
        cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)

        #sample sizes
        sample_size_best_in_class, sample_size_gen_error = 1e4

        #generate data sets
        data_gen_error = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                         mu=mu, sample_size=sample_size_best_in_class, seed = 21)
        data_best_in_class = generate_data.generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                             mu=mu, sample_size=sample_size_best_in_class, seed = 7)

        #fix the arguments except for sample_size and then use map to vectorize over sample size values
        compute_sample_risk_gen_error_partial = functools.partial(compute_sample_risk_gen_error,
                                                data_best_in_class = data_best_in_class,
                                                data_gen_error = data_gen_error,
                                                poly_order_range = range(1, 5))

        #compute estimation error for each sample size between 5 and 1000 (inclusive)
        # Call the partial function with est_sample_sizes
        est_sample_sizes = range(5, 1000)
        est_gen_errors = list(map(compute_sample_risk_gen_error_partial, est_sample_sizes))

        #Create data frame for estimation error
        pd_est_error = pd.DataFrame({"sample_size": np.array(est_sample_sizes),
                                     "sample_risk_gen_error": est_gen_errors})

    """
    with resources.path("smltheory.data", "data_est_error.csv") as path:
        data_est_error = pd.read_csv(path)

    return data_est_error


def get_data_opt_error():
    """Returns data set for depicting optimization error.

    Returns
    -------
    data_opt_error:  pandas.core.frame.DataFrame
        Data set for depicting optimization error

    Notes
    -----
   Data set for depicting optimization error was generated using the following code:

    .. code-block:: python

        sd = [1.5, 1.2]
        mu = [5, 7]
        sd = [1.2, 1.7]
        rho_weather_winemaking = 0.35
        cov_matrix = generate_data.create_covariance_matrix(sd=sd, rho=rho_weather_winemaking)

        #sample sizes
        sample_size_best_in_class, sample_size_gen_error = 1e4

        #generate data sets
        data_gen_error = generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                         mu=mu, sample_size=sample_size_best_in_class, seed = 21)
        data_best_in_class = generate_mult_trunc_normal(cov_matrix=cov_matrix, sd=sd,
                             mu=mu, sample_size=sample_size_best_in_class, seed = 7)

        #fix the arguments except for sample_size and then use map to vectorize over sample size values
        get_opt_risk_min_partial = functools.partial(get_opt_risk_min,
                                                     data_best_in_class = data_best_in_class,
                                                     data_gen_error = data_gen_error)

        # Call the partial function with each sample size specified in opt_sample_sizes
        opt_sample_sizes = range(5, 1000)
        with concurrent.futures.ThreadPoolExecutor(max_workers = 3) as executor:
            opt_gen_errors = list(executor.map(get_opt_risk_min_partial, opt_sample_sizes))

        #Create data frame
        df_opt_error = pd.DataFrame({"sample_size": np.array(opt_sample_sizes),
                                     "opt_gen_error": opt_gen_errors})

    """
    with resources.path("smltheory.data", "data_opt_error_fast.csv") as path:
        data_opt_error = pd.read_csv(path)

    return data_opt_error


def get_data_bv_tradeoff():
    """Returns data set for depicting bias-variance tradeoff.

    Returns
    -------
    data_bv_tradeoff:  pandas.core.frame.DataFrame
        Data set for depicting bias-variance tradeoff.

    Notes
    -----
    Given that code for computing this data set was more complicated, see see
    whitepaper at `sebastiansciarra.com <https://sebastiansciarra.com/technical_content/understanding_ML>`_.
    """
    with resources.path("smltheory.data", "data_bv_tradeoff.csv") as path:
        data_bv_tradeoff = pd.read_csv(path)

    return data_bv_tradeoff



# read version from installed package
from importlib.metadata import version

__version__ = version("smltheory")

from smltheory.generate_data import *
from smltheory.overfitting import *
from smltheory.opt_error import *
from smltheory.est_error import *

from smltheory.least_squares import *
from smltheory.gradient_descent import *
from smltheory.trun_mvnt import *
from smltheory.bayes_decision import *

from smltheory.datasets import *

